/*
 * Copyright (c) 2004 Damir Bikmuhametov <boco@ufanet.ru>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: avecomm.c,v 1.9 2004/03/14 10:56:54 boco Exp $
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>

#include "common.h"
#include "avecomm.h"

static char * extract_virus_name(char *);

/* main scanning procedure */

int ave_scan_file(char * filename, char ** viruslist)
{
	char * id = "ave_scan_file";
	char buffer[4096];
	int fd = -1;
	int len;
	int infected = 0;
	int end_of_conversation = 0;
	char * p, * q;
	char * virus_names[11];
	char ** virus_name;
	int viruses = 1;
	int num_tries = 3;	/* will do 3 attemts to connect to aveserver */

	if (!filename || !*filename) {
		syslog(LOG_ERR, "%s: empty filename", id);
		return -1;
	}

	if (!av_engine_socket || !*av_engine_socket) {
		syslog(LOG_ERR, "%s: empty AVE socket name", id);
		return -1;
	}

	while (num_tries--) {

		/* connect to aveserver */

		if ((fd = av_engine_connect(av_engine_socket)) == -1) {
			syslog(LOG_ERR, "%s: error connecting to AVE socket: %m", id);
			return -1;
		}

		memset(buffer, 0, sizeof(buffer));

		if ((len = timed_read(fd, buffer, sizeof(buffer) - 1, av_engine_timeout)) == -1 ) {
			syslog(LOG_ERR, "%s: can't read AVE banner: %m", id);
			av_engine_disconnect(fd);
			return -1;
		}

		p = buffer;

		if (!(q = getline(&p))) {
			syslog(LOG_ERR, "%s: Can't get AVE banner", id);
			av_engine_disconnect(fd);
			fd = -1;
			num_tries = 0;
		}

		switch (*q) {
			case '2':	/* successfully connected */
				num_tries = 0;
				break;
			case '4':	/* AV engine is busy */
				syslog(LOG_ERR, "%s: AVE is busy, will try %d times more. AVE banner: %s", id, num_tries, q);
				av_engine_disconnect(fd);
				fd = -1;
				sleep(3);	/* pause for 3 seconds */
					        /* FIXME: should we use random(3)? */
				break;
			default:	/* any 5xx code, get out */
				syslog(LOG_ERR, "%s: invalid AVE banner: %s", id, q);
				av_engine_disconnect(fd);
				fd = -1;
				num_tries = 0;
		}	/* switch() */

	}	/* while() */

	if (fd == -1)
		return -1; /* can't setup connection */

	/* prepare array for virus names */

	memset(virus_names, 0, sizeof(virus_names));
	virus_name = virus_names;

	/* prepare AVE command */

	snprintf(buffer, sizeof(buffer) - 1, "SCAN xmPQRSTUWbgh %s\r\n",
		       filename);

	len = strlen(buffer);
	
	if (timed_write(fd, buffer, len, av_engine_timeout) != len) {
		syslog(LOG_ERR, "%s: can't write to AVE socket: %m", id);
		av_engine_disconnect(fd);
		return -1;
	}

	while (!end_of_conversation && (len = timed_read(fd, buffer, sizeof(buffer) - 1, av_engine_timeout))) {

		if (len == -1) {
			syslog(LOG_ERR, "%s: can't read from AVE socket: %m",
				       id);
			FREE_ARRAY(virus_names);
			av_engine_disconnect(fd);
			return -1;
		}

		buffer[len] = 0;
		p = buffer;
		while ((q = getline(&p))) {
			if (*q == '5') {
				syslog(LOG_ERR, "%s: error from AVE: %s",
					       id, q);
				infected = -1;
				end_of_conversation++;
				break;
			} else if (*q == '2') {
				q[3] = 0;
				switch (atoi(q)) {
					case 221:
					case 230:
					case 231:
					case 232:
					case 233:	infected = 1;
				}
				end_of_conversation++;
				break;
			} else if (viruses < ARRAY_SIZE(virus_names) &&
					!memcmp(q, "32", 2) &&
					(q[2] == '2' ||
					 q[2] == '7' ||
					 q[2] == '8'
				       	)) {
				*virus_name = extract_virus_name(q);
				if (*virus_name) {
					virus_name++;
					viruses++;
				}
			}
		}
	}
	
	strcpy(buffer, "QUIT\r\n");

	len = strlen(buffer);

	if (timed_write(fd, buffer, len, av_engine_timeout) != len) {
		syslog(LOG_ERR, "%s: can't write to AVE socket: %m", id);
		FREE_ARRAY(virus_names);
		av_engine_disconnect(fd);
		return -1;
	}

	memset(buffer, 0, sizeof(buffer));

	if ((len = timed_read(fd, buffer, sizeof(buffer) - 1, av_engine_timeout)) == -1) {
		syslog(LOG_ERR, "%s: can't read from AVE socket: %m", id);
		FREE_ARRAY(virus_names);
		av_engine_disconnect(fd);
		return -1;
	}

	/* it doesn't matter what AVE said at exit so we don't analyze it */

	av_engine_disconnect(fd);

	if (infected == 1) {
		len = 0;
		virus_name = virus_names;
		while (*virus_name) {
			len += strlen(*virus_name) + 1;
			virus_name++;
		}
		len++;
		if ((*viruslist = (char *) malloc(len))) {
			memset(*viruslist, 0, len);
			virus_name = virus_names;
			while (*virus_name) {
				strcat(*viruslist, *virus_name);
				strcat(*viruslist, " ");
				virus_name++;
			}
		}
	}

	FREE_ARRAY(virus_names);

	return infected;
}

/* this routine gets string returned by AVE and returns virus name */

static char * extract_virus_name(char * line)
{
	char * p;
	int len;
	char * name = NULL;

	line += 4; /* now line points to virus name */
	if ((p = strchr(line, ' '))) {
		len = p - line;
		if ((name = (char *) malloc(len + 1))) {
			memset(name, 0, len + 1);
			strncat(name, line, len);
		}
	}
	return name;
}
